"""Test schema validation."""

from spintest.validator import input_validator, TASK_SCHEMA


def test_TASK_SCHEMA():
    right_input = {"method": "GET", "route": "/foo/bar"}
    wrong_input = {"foo": "bar"}

    assert input_validator(right_input, TASK_SCHEMA)
    assert not input_validator(wrong_input, TASK_SCHEMA)


def test_task_schema_with_optional_key_expected_match_set_to_strict():
    right_input = {
        "method": "string",
        "route": "string",
        "expected": {"expected_match": "strict"},
    }

    assert input_validator(right_input, TASK_SCHEMA)


def test_task_schema_with_optional_key_expected_match_set_to_partial():
    right_input = {
        "method": "string",
        "route": "string",
        "expected": {"expected_match": "partial"},
    }

    assert input_validator(right_input, TASK_SCHEMA)


def test_task_schema_with_optional_key_expected_match_set_to_dummy_value():
    right_input = {
        "method": "string",
        "route": "string",
        "expected": {"expected_match": "dummy_value"},
    }

    assert None is input_validator(right_input, TASK_SCHEMA)



"""Task representation."""

import asyncio

import jinja2
import json
import requests
import time

from urllib.parse import urljoin

from spintest import logger
from spintest.validator import input_validator, TASK_SCHEMA
from spintest.types import type_aware_encoder


class Task(object):
    """Task handler."""

    def __init__(self, url: str, task: dict, output: dict, verify: bool = True):
        """Initialization of `Task` class."""
        self.url = url
        self.task = task
        self.rollback = self.task.pop("rollback", None)
        self.output = output
        self.verify = verify
        self.response = None

    def _response(self, status: str, message: str) -> dict:
        """Return the response with logging."""
        result = {
            "name": self.task.get("name"),
            "status": status,
            "timestamp": time.asctime(),
            "url": self.url,
            "route": self.task.get("route", "/"),
            "message": message,
            "code": self._response_code(),
            "body": self._response_body(),
            "task": self.task,
            "ignore": self.task.get("ignore", False),
        }
        if "headers" in self.task and "Authorization" in self.task["headers"]:
            self.task["headers"]["Authorization"] = "****"

        log_level = {"SUCCESS": logger.info, "FAILED": logger.error}
        log_level.get(status, logger.critical)(json.dumps(result, indent=4))

        result["output"] = self.output
        return result

    def _response_code(self):
        """Response code formatter."""
        try:
            return self.response.status_code
        except AttributeError:
            return None

    def _response_body(self):
        """Response body formatter."""
        try:
            return self.response.json()
        except ValueError:
            return self.response.text
        except AttributeError:
            return None

    def validate_code(self):
        """Validate the returned status code."""
        expected_code = self.task.get("expected", {}).get("code")
        response_code = self._response_code()
        if expected_code and expected_code != response_code:
            return self._response("FAILED", "Invalid HTTP status code.")

        if not (expected_code or 200 <= response_code < 300):
            return self._response("FAILED", "Invalid default HTTP status code (2XX).")

    def _expected_match(self):
        return self.task.get("expected", {}).get("expected_match", "strict")

    def _compare_body(self, body, expected):
        """Recursive comparison of body.
        if expected value is None, any body is accepted.
        if expected value is set, and expect_match value is not set,
            strict comparison is made.
        if expected value is set, and expect_match value is set to "partial",
            partial comparison is made.
        if expected and body keys are different, returns false.
        """
        if expected is None:
            return True

        if not isinstance(body, type(expected)):
            return False

        if isinstance(body, dict):
            if self._expected_match() == "strict" and body.keys() != expected.keys():
                return False

            if not set(expected).issubset(body):
                return False

            return all(
                [self._compare_body(body[ek], expected[ek]) for ek in expected.keys()]
            )

        elif isinstance(body, list):
            if self._expected_match() == "strict":
                if len(body) != len(expected):
                    return False
                for body_item in body:
                    for expected_item in expected:
                        if self._compare_body(body_item, expected_item):
                            break
                    else:
                        return False
            else:
                for expected_item in expected:
                    for body_item in body:
                        if self._compare_body(body_item, expected_item):
                            break
                    else:
                        return False
            return True

        else:
            if body == expected:
                return True
            return False

    def validate_body(self):
        """Validate the returned body."""
        expected_body = self.task.get("expected", {}).get("body")
        response_body = self._response_body()
        if expected_body and not self._compare_body(response_body, expected_body):
            return self._response(
                "FAILED",
                "The response body does not correspond with the expected body.",
            )

    async def run(self) -> dict:
        """Run the task on a specified URL."""

        # -- Input validation --

        validated_task = input_validator(self.task, TASK_SCHEMA)
        if not validated_task:
            return self._response(
                "FAILED", f"Task must follow this schema : {TASK_SCHEMA}."
            )

        self.task = validated_task

        self.task["method"] = self.task["method"].upper()
        if self.task["method"] not in [
            "GET",
            "POST",
            "PATCH",
            "PUT",
            "DELETE",
            "HEAD",
            "CONNECT",
            "OPTIONS",
            "TRACE",
        ]:
            return self._response("FAILED", "Invalid HTTP method.")

        # Jinja2 logic substitution
        template = jinja2.Template(
            json.dumps(self.task, cls=type_aware_encoder(self.output))
        )
        self.task = json.loads(template.render(**self.output))

        self.task["headers"] = {
            **{"Accept": "application/json", "Content-Type": "application/json"},
            **self.task.get("headers", {}),
        }

        # -- Request --

        loop = asyncio.get_event_loop()

        for _ in range(self.task["retry"] + 1):
            try:
                if self.output.get("__token__"):
                    token = self.output["__token__"]
                    self.task["headers"]["Authorization"] = "Bearer " + (
                        token() if callable(token) else token
                    )
                self.response = await loop.run_in_executor(
                    None,
                    lambda: getattr(requests, self.task["method"].lower())(
                        urljoin(self.url, self.task["route"]),
                        json=self.task.get("body"),
                        headers=self.task["headers"],
                        verify=self.verify,
                    ),
                )
            except requests.exceptions.RequestException:
                failed_response = self._response("FAILED", "Request failed.")
                await asyncio.sleep(self.task["delay"])
                continue

            # -- Output validation --

            failed_response = self.validate_code()
            if failed_response is not None:
                await asyncio.sleep(self.task["delay"])
                continue

            failed_response = self.validate_body()
            if failed_response is not None:
                await asyncio.sleep(self.task["delay"])
                continue

            output_variable = self.task.get("output")
            if output_variable:
                self.output[output_variable] = self._response_body()

            return self._response("SUCCESS", "OK.")

        return failed_response



import json
import httpretty

from spintest import logger, spintest
from spintest.types import Int, List, Float, Bool

logger.disabled = True


def test_task_with_custom_type_evaluation():
    """Test spintest with a strict match on custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps(
            {"inner": {"foo": 2, "bar": "buz", "qux": 2.23, "foobar": True}}
        ),
        status=200,
    )
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=200)

    spintest(
        ["http://test.com"],
        [
            {"method": "GET", "route": "/test", "output": "test"},
            {
                "method": "POST",
                "route": "/test",
                "body": {
                    "int": Int("{{ test['inner']['foo'] }}"),
                    "list": List("{{ test['inner']['bar'] }}"),
                    "float": Float("{{ test['inner']['qux'] }}"),
                    "bool": Bool("{{ test['inner']['foobar'] }}"),
                },
            },
        ],
    )

    assert json.loads(httpretty.last_request().body) == {
        "int": 2,
        "list": ["b", "u", "z"],
        "float": 2.23,
        "bool": True,
    }

    httpretty.disable()
    httpretty.reset()



import asyncio
import json
import uuid
import httpretty
import pytest
from spintest import logger, spintest, TaskManager

logger.disabled = True


class FakeCreatedToken:
    """Function use to create token."""

    def __init__(self):
        self.call_count = 0

    def __call__(self):
        self.call_count += 1
        return str(self.call_count)


def test_task_with_token_token_is_not_visible_on_log():
    """Test spintest with a strict match on custom body."""
    TOKEN = str(uuid.uuid4())
    httpretty.enable()

    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )

    tasks = [
        {
            "method": "GET",
            "route": "/test",
            "expected": {
                "code": 200,
                "body": {"a": "a", "b": "b", "c": "c"},
                "expected_match": "strict",
            },
        }
    ]

    loop = asyncio.get_event_loop()
    results = []
    manager = TaskManager(["http://test.com"], tasks, token=TOKEN)

    for i in range(len(tasks)):
        result = loop.run_until_complete(manager.next())
        results.append(result)

    assert TOKEN not in results[0]["task"]["headers"]["Authorization"]

    httpretty.disable()
    httpretty.reset()


def test_task_with_fix_token_set_configuration():
    """Test spintest with a strict match on custom body."""
    httpretty.enable()

    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"a": "a", "b": "b", "c": "c"},
                    "expected_match": "strict",
                },
            }
        ],
        token="ABC",
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_with_fuct_token_set_configuration():
    """Test spintest with a strict match on custom body."""
    httpretty.enable()

    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )

    token = FakeCreatedToken()
    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"a": "a", "b": "b", "c": "c"},
                    "expected_match": "strict",
                },
            }
        ],
        token=token(),
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_two_tokens_are_generated_with_one_task_with_two_retry():
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        status=200,
        body=json.dumps({"status": "CREATED"}),
    )
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        status=200,
        body=json.dumps({"status": "CREATING"}),
    )
    loop = asyncio.new_event_loop()
    token_func = FakeCreatedToken()
    manager = TaskManager(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {"body": {"status": "CREATED"}},
                "retry": 1,
            }
        ],
        token=token_func,
    )
    result = loop.run_until_complete(manager.next())
    assert "SUCCESS" == result["status"]
    assert token_func.call_count == 2

    httpretty.disable()
    httpretty.reset()


def test_task_with_ayncio_loop_and_fuct_token_set_configuration():
    """Test spintest with a strict match on custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )
    loop = asyncio.new_event_loop()
    token_func = FakeCreatedToken()
    manager = TaskManager(
        ["http://test.com"],
        [{"method": "GET", "route": "/test"}, {"method": "GET", "route": "/test"}],
        token=token_func,
    )
    result = loop.run_until_complete(manager.next())
    assert "SUCCESS" == result["status"]
    assert token_func.call_count == 1
    result1 = loop.run_until_complete(manager.next())
    assert "SUCCESS" == result1["status"]
    assert token_func.call_count == 2

    httpretty.disable()
    httpretty.reset()


def test_task_strict_all_keys():
    """Test spintest with a strict match on custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"a": "a", "b": "b", "c": "c"},
                    "expected_match": "strict",
                },
            }
        ],
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_partial_value_works_like_strict_keys_values():
    """Test spintest with a partial custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"a": "a", "b": "b", "c": "c"},
                    "expected_match": "partial",
                },
            }
        ],
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_strict_expected_complete_list_values():
    """Test spintest with a strict match on body with list."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"keys": ["a", "b", "c"]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"keys": ["a", "b", "c"]},
                    "expected_match": "strict",
                },
            }
        ],
    )
    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_strict_expected_cmp_list_with_multiple_identics_values():
    """Test spintest with a strict match on custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"keys": ["a", "b", "c"]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"keys": ["a", "b", "b"]},
                    "expected_match": "strict",
                },
            }
        ],
    )
    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_strict_expected_cmp_list_with_many_identics_double_values():
    """Test spintest with a strict match on custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"keys": ["a", "b", "b"]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"keys": ["a", "b", "a", "a"]},
                    "expected_match": "strict",
                },
            }
        ],
    )
    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_strict_expectedlist_multiple_None():
    """Test spintest with a None value in custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"keys": ["a", "b", "c"]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"keys": [None, "a", None]},
                    "expected_match": "strict",
                },
            }
        ],
    )
    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_strict_expected_list_None():
    """Test spintest with a None value in custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"keys": ["a", "b"]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"keys": [None, "a"]},
                    "expected_match": "strict",
                },
            }
        ],
    )
    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_stric_expected_partial_list():
    """Test spintest with a None value in custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"keys": ["a", "b", "c"]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"keys": ["a"]},
                    "expected_match": "strict",
                },
            }
        ],
    )
    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_partial_expected_list():
    """Test spintest with a partial match on custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"keys": ["a", "b", "c"]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"keys": ["a"]},
                    "expected_match": "partial",
                },
            }
        ],
    )
    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_partial_expected_dict_keys():
    """Test spintest with a partial match on a custom body with dict."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"b": "b"},
                    "expected_match": "partial",
                },
            }
        ],
    )
    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_partial_expected_dict_keys_failed():
    """Test spintest with a failed partial match on a custom body with dict."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "code": 200,
                    "body": {"d": "d"},
                    "expected_match": "partial",
                },
            }
        ],
    )
    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_basic():
    """Test basic spintest."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )

    result = spintest(["http://test.com"], [{"method": "GET", "route": "/test"}])

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_basic_no_verify():
    """Test basic spintest with verify set to false."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )

    result = spintest(
        ["http://test.com"], [{"method": "GET", "route": "/test"}], verify=False
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_basic_parallel():
    """Test spintest with parallel urls."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )

    result = spintest(
        ["http://test.com"], [{"method": "GET", "route": "/test"}], parallel=True
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_basic_token():
    """Test spintest with basic token."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )

    result = spintest(
        ["http://test.com"], [{"method": "GET", "route": "/test"}], token="string"
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_invalid_method():
    """Test spintest with invalid method."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )

    result = spintest(["http://test.com"], [{"method": "FOO", "route": "/test"}])

    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_basic_templating():
    """Test spintest with templating."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )
    httpretty.register_uri(httpretty.GET, "http://test.com/bar")

    result = spintest(
        ["http://test.com"],
        [
            {"method": "GET", "route": "/test", "output": "test"},
            {"method": "GET", "route": "/{{ test['foo'] }}"},
        ],
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_invalid_task_no_method():
    """Test spintest with an invalid task."""
    httpretty.enable()
    httpretty.register_uri(httpretty.GET, "http://test.com/test")

    result = spintest(["http://test.com"], [{"route": "/test"}])

    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_invalid_task_default_route():
    """Test spintest with an invalid task."""
    httpretty.enable()
    httpretty.register_uri(httpretty.GET, "http://test.com/")

    result = spintest(["http://test.com"], [{"method": "GET"}])

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_bad_status_code():
    """Test spintest against an url with 404 status code."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"error": "foo"}),
        status=404,
    )

    result = spintest(["http://test.com"], [{"method": "GET", "route": "/test"}])

    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_custom_status_code():
    """Test spintest with a custom statius code."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"error": "foo"}),
        status=404,
    )

    result = spintest(
        ["http://test.com"],
        [{"method": "GET", "route": "/test", "expected": {"code": 404}}],
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_custom_status_code_failed():
    """Test spintest with a wrong custom status code."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"error": "foo"}),
        status=404,
    )

    result = spintest(
        ["http://test.com"],
        [{"method": "GET", "route": "/test", "expected": {"code": 403}}],
    )

    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_custom_body():
    """Test spintest with a custom body and status code."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"error": "foo"}),
        status=404,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {"code": 404, "body": {"error": "foo"}},
            }
        ],
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_partial_body():
    """Test spintest with a partial custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps(
            {
                "a": "a",
                "b": 1,
                "c": "c",
                "d": {"a": "a", "b": 1, "c": "c"},
                "e": {"a": "a", "b": 1, "c": "c"},
                "f": [{"a": "a", "b": 1, "c": "c"}, {"d": "d", "e": 2, "f": "f"}],
            }
        ),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "body": {
                        "a": "a",
                        "b": 1,
                        "c": None,
                        "d": {"a": "a", "b": 1, "c": None},
                        "e": None,
                        "f": [
                            {"a": None, "b": None, "c": None},
                            {"d": "d", "e": 2, "f": None},
                        ],
                    }
                },
            }
        ],
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_partial_response_body_validation_with_expected_key_name_with_needed_param():
    """Test spintest can validate a response body with an expected partial
    body.
    """
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": "a", "b": "b", "c": "c"}),
        status=200,
    )
    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {
                    "body": {"a": "a", "b": "b", "c": "c"},
                    "expected_match": "strict",
                },
            }
        ],
    )
    assert True is result
    httpretty.disable()
    httpretty.reset()


def test_task_partial_body_invalid_type():
    """Test spintest with a failed partial custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"a": "a"}), status=200
    )

    result = spintest(
        ["http://test.com"],
        [{"method": "GET", "route": "/test", "expected": {"body": {"a": 1}}}],
    )

    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_partial_body_invalid_list():
    """Test spintest with a failed partial custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"a": [{"a": "a"}, {"b": "b"}]}),
        status=200,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {"body": {"a": [{"c": "c"}]}},
            }
        ],
    )

    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_custom_body_flat():
    """Test spintest with a flat body."""
    httpretty.enable()
    httpretty.register_uri(httpretty.GET, "http://test.com/test", body="text")

    result = spintest(
        ["http://test.com"],
        [{"method": "GET", "route": "/test", "expected": {"body": "text"}}],
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_custom_body_failed():
    """Test spintest with a wrong custom body."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"error": "foo"}),
        status=404,
    )

    result = spintest(
        ["http://test.com"],
        [
            {
                "method": "GET",
                "route": "/test",
                "expected": {"code": 404, "body": {"error": "bar"}},
            }
        ],
    )

    assert False is result

    httpretty.disable()
    httpretty.reset()


def test_task_next():
    """Test spintest accessing with next."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.GET, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )

    loop = asyncio.new_event_loop()
    manager = TaskManager(["http://test.com"], [{"method": "GET", "route": "/test"}])
    result = loop.run_until_complete(manager.next())

    assert "SUCCESS" == result["status"]
    assert "OK." == result["message"]
    assert 200 == result["code"]
    assert {"foo": "bar"} == result["body"]
    assert "http://test.com" == result["url"]
    assert "/test" == result["route"]
    assert {
        "method": "GET",
        "route": "/test",
        "headers": {"Accept": "application/json", "Content-Type": "application/json"},
        "retry": 0,
        "delay": 1,
        "ignore": False,
    } == result["task"]

    assert {"__token__": None} == result["output"]

    with pytest.raises(StopAsyncIteration):
        loop.run_until_complete(manager.next())

    httpretty.disable()
    httpretty.reset()


def test_task_next_async():
    """Test spintest accessing with next asynchronously."""
    httpretty.enable()
    httpretty.register_uri(httpretty.GET, "http://foo.com/test")
    httpretty.register_uri(httpretty.GET, "http://bar.com/test")

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://foo.com", "http://bar.com"],
        [{"method": "GET", "route": "/test"}],
        parallel=True,
    )
    results = loop.run_until_complete(manager.next())

    for result in results:
        assert "SUCCESS" == result["status"]

    with pytest.raises(StopAsyncIteration):
        loop.run_until_complete(manager.next())

    httpretty.disable()
    httpretty.reset()


def test_task_next_async_failed():
    """Test spintest accessing with next asynchronously."""
    httpretty.enable()
    httpretty.register_uri(httpretty.GET, "http://foo.com/test")
    httpretty.register_uri(httpretty.GET, "http://bar.com/test", status=500)

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://foo.com", "http://bar.com"],
        [{"method": "GET", "route": "/test"}],
        parallel=True,
    )
    results = loop.run_until_complete(manager.next())

    assert "SUCCESS" in [result["status"] for result in results]
    assert "FAILED" in [result["status"] for result in results]

    with pytest.raises(StopAsyncIteration):
        loop.run_until_complete(manager.next())

    httpretty.disable()
    httpretty.reset()


def test_task_with_ignore_option():
    """Test spintest with ignore option."""
    httpretty.enable()
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)

    result = spintest(
        ["http://test.com"], [{"method": "GET", "route": "/test", "ignore": True}]
    )

    assert True is result

    httpretty.disable()
    httpretty.reset()


def test_task_with_ignore_option_followed_by_error():
    """Test spintest with ignore option."""
    httpretty.enable()
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)

    result = spintest(
        ["http://test.com"],
        [
            {"method": "GET", "route": "/test", "ignore": True},
            {"method": "GET", "route": "/test"},
        ],
    )

    assert False is result

    httpretty.disable()
    httpretty.reset()



"""Allow computed value to have types."""

import jinja2
import json


class JSONValue(object):
    """A base class that implements type-aware serialization values."""

    def __init__(self, value):
        self.value = value

    def get_value(self, output):
        template = jinja2.Template(self.value)
        return template.render(**output)

    def serialize(self, output):
        return self.get_value(output)


class Int(JSONValue):
    """Converts value to `int`."""

    def serialize(self, output):
        return int(self.get_value(output))


class List(JSONValue):
    """Converts value to `list`."""

    def serialize(self, output):
        return list(self.get_value(output))


class Float(JSONValue):
    """Converts value to `float`."""

    def serialize(self, output):
        return float(self.get_value(output))


class Bool(JSONValue):
    """Converts value to `bool`."""

    def serialize(self, output):
        return bool(self.get_value(output))


class TypeAwareEncoder(json.JSONEncoder):

    output = None

    def default(self, obj):
        if isinstance(obj, JSONValue):
            return obj.serialize(self.output)
        return json.JSONEncoder.default(self, obj)


def type_aware_encoder(output):
    TypeAwareEncoder.output = output
    return TypeAwareEncoder



"""Input validation."""

import typing

from schema import Schema, SchemaError, Or, Optional


TASK_SCHEMA = Schema(
    {
        "method": str,
        Optional("route", default="/"): str,
        Optional("name"): str,
        Optional("body"): dict,
        Optional(
            "headers",
            default={"Accept": "application/json", "Content-Type": "application/json"},
        ): dict,
        Optional("output"): str,
        Optional("expected"): {
            Optional("code"): int,
            Optional("body"): Or(dict, str),
            Optional("expected_match", default="strict"): Or("partial", "strict"),
        },
        Optional("retry", default=0): int,
        Optional("delay", default=1): int,
        Optional("ignore", default=False): bool,
        Optional("rollback"): [Or(str, dict)],
    }
)


def input_validator(input, input_schema) -> typing.Optional[dict]:
    try:
        return input_schema.validate(input)
    except SchemaError:
        return None



"""Initialization of `spintest` package."""

import asyncio
import colorlog
import logging

from typing import List, Dict


handler = colorlog.StreamHandler()
handler.setFormatter(
    colorlog.ColoredFormatter("%(log_color)s%(asctime)s - %(levelname)s - %(message)s")
)
logger = colorlog.getLogger("spintest")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


from spintest.manager import TaskManager  # noqa: E402


def spintest(
    urls: List[str],
    tasks: List[Dict[str, str]],
    token: str = None,
    parallel: bool = False,
    verify: bool = True,
):
    """Programmatic wrapper for spintest."""
    loop = asyncio.new_event_loop()
    task_manager = TaskManager(
        urls, tasks, token=token, parallel=parallel, verify=verify
    )
    result = loop.run_until_complete(task_manager.run())
    loop.close()
    return result


__all__ = ["spintest", "TaskManager"]
__version__ = "0.2.0"



"""Setup for `spintest` package."""

from setuptools import setup, find_packages


def parse_requirements(requirements_file):
    with open(requirements_file) as rf:
        requirements = rf.read()
    return requirements


setup(
    name="spintest",
    version="0.2.0",
    license="BSD-3-Clause",
    author="Matthieu Gouel",
    author_email="matthieu.gouel@gmail.com",
    url="https://github.com/societe-generale/spintest",
    description="Functional scenario interpreter",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Topic :: Utilities",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
    ],
    include_package_data=True,
    packages=find_packages(),
    install_requires=parse_requirements("requirements.txt"),
)



"""Task Manager representation."""

import asyncio
import itertools

from typing import Callable, Dict, List, Union

from spintest import logger
from spintest.task import Task


class TaskManager(object):
    """Task manager."""

    def __init__(
        self,
        urls: List[str],
        tasks: List[Dict[str, str]],
        token: Union[str, Callable[..., str]] = None,
        parallel: bool = False,
        verify: bool = True,
    ):
        """Initialization of `TaskManager` class."""
        self.urls = urls
        self.tasks = tasks
        self.rollback_tasks = []
        self.token = token
        self.verify = verify
        self.parallel = parallel

        if self.parallel:
            self.outputs = [{"__token__": self.token}] * len(self.urls)
            self.stack = self._parallel_executor()
        else:
            self.outputs = [{"__token__": self.token}]
            self.stack = self._executor()

    @staticmethod
    def _error(critical: str = None):
        """Throw a custom error."""
        if critical:
            logger.critical(critical)
        return [{"status": "FAILED", "ignore": False}]

    def validate_refs(self) -> bool:
        """Validate the integrity of task references."""
        task_names = [task["name"] for task in self.tasks if task.get("name")]
        for task in self.tasks:
            for rollback in task.get("rollback", []):
                if isinstance(rollback, str) and rollback not in task_names:
                    logger.critical("Reference validation failed.")
                    return False
        return True

    def rollback_lookup(self, name, activate_ignore=True):
        """Get a task from a name."""
        for task in self.tasks:
            if task.get("name") == name:
                rollback_task = task.copy()
                if activate_ignore:
                    rollback_task["ignore"] = True
                return rollback_task

    def rollback_register(self, url, task):
        """Register rollback tasks."""
        for rollback in task.get("rollback", [])[::-1]:
            if isinstance(rollback, str):
                self.rollback_tasks.append((url, self.rollback_lookup(rollback)))
            elif isinstance(rollback, dict):
                rollback["ignore"] = True
                self.rollback_tasks.append((url, rollback))
            else:
                return False
        return True

    async def rollback_executor(self, url=None):
        """Execute the rollback stack."""
        for rollback_url, rollback_task in self.rollback_tasks[::-1]:
            if url is not None and rollback_url != url:
                continue

            if self.parallel:
                output = self.outputs[self.urls.index(url)].copy()
            else:
                output = self.outputs[0].copy()

            result = await Task(
                rollback_url, rollback_task, output=output, verify=self.verify
            ).run()

            if self.parallel:
                self.outputs[self.urls.index(url)] = result["output"]
            else:
                self.outputs = [result["output"]]

            yield [result]

    async def _executor(self) -> list:
        """Private task executor."""
        is_refs_validated = self.validate_refs()
        if not is_refs_validated:
            yield self._error(critical="Rollback scenario validation failed.")
            return

        for url in self.urls:
            is_success = True
            for task in self.tasks:
                is_registered = self.rollback_register(url, task)
                if not is_registered:
                    yield self._error(critical="Invalid rollback schema.")
                    return

                result = await Task(
                    url, task, output=self.outputs[0].copy(), verify=self.verify
                ).run()

                self.outputs = [result["output"]]

                yield [result]
                if result["status"] != "SUCCESS" and result["ignore"] is False:
                    is_success = False
                    break

            if not is_success:
                async for rollback in self.rollback_executor():
                    yield rollback

    async def _parallel_executor(self) -> list:
        """Private parallel task executor."""
        is_refs_validated = self.validate_refs()
        if not is_refs_validated:
            yield self._error(critical="Rollback scenario validation failed.")
            return

        state = {url: None for url in self.urls}
        for task in self.tasks:
            task_run_list = []
            for i, url in enumerate(self.urls):
                if (
                    state[url] is not None
                    and state[url]["status"] != "SUCCESS"
                    and state[url]["ignore"] is False
                ):
                    continue

                is_registered = self.rollback_register(url, task)
                if not is_registered:
                    yield self._error(critical="Invalid rollback schema.")
                    return

                task_run_list.append(
                    Task(
                        url, task, output=self.outputs[i].copy(), verify=self.verify
                    ).run()
                )

            results = await asyncio.gather(*task_run_list)

            for url in state:
                for result in results:
                    if url == result["url"]:
                        self.outputs[self.urls.index(url)] = result["output"]
                        state[url] = result
                        break

            yield results

        for url in self.urls:
            if state[url]["status"] != "SUCCESS" and state[url]["ignore"] is False:
                async for rollback in self.rollback_executor(url=url):
                    yield rollback

    async def _next(self) -> list:
        """Execute the next task."""
        return await self.stack.__anext__()

    async def next(self) -> Union[str, list]:
        """Wrapper for better iterative output."""
        result = await self._next()
        if len(result) == 1:
            return result[0]
        else:
            return result

    async def run(self) -> bool:
        """Run the whole task queue."""
        results = []
        while True:
            try:
                results.append(await self._next())
            except StopAsyncIteration:
                break

        return all(
            [
                result["status"] == "SUCCESS"
                for result in list(itertools.chain.from_iterable(results))
                if result["ignore"] is False
            ]
        )



"""Test of rollback feature."""

import asyncio
import httpretty
import json

from spintest import logger, spintest, TaskManager

logger.disabled = True


def test_rollback_unkwnown_ref():
    """Test spintest with unknown rollback reference."""
    result = spintest(
        ["http://test.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"method": "DELETE", "route": "/test"},
        ],
    )

    assert result is False


def test_rollback_unkwnown_ref_parallel():
    """Test spintest with unknown rollback reference parallel."""
    result = spintest(
        ["http://test.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"method": "DELETE", "route": "/test"},
        ],
        parallel=True,
    )

    assert result is False


def test_rollback_invalid_schema():
    """Test spintest with unknown rollback reference."""
    result = spintest(
        ["http://test.com"],
        [
            {"method": "POST", "route": "/test", "rollback": [["toto"]]},
            {"method": "GET", "route": "/test"},
            {"method": "DELETE", "route": "/test"},
        ],
    )

    assert result is False


def test_rollback_ref():
    """Test spintest with rollback reference."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test", status=204)

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://test.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
        ],
    )

    for _ in range(3):
        result = loop.run_until_complete(manager.next())

    assert result["name"] == "delete_test"
    assert result["ignore"] is True

    httpretty.disable()
    httpretty.reset()


def test_rollback_ref_parallel():
    """Test spintest with rollback parallel."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test", status=204)

    spintest(
        ["http://test.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
        ],
        parallel=True,
    )

    assert httpretty.last_request().method == httpretty.DELETE

    httpretty.disable()
    httpretty.reset()


def test_rollback_task():
    """Test spintest with rollback task."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test", status=204)

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://test.com"],
        [
            {
                "method": "POST",
                "route": "/test",
                "rollback": [
                    {"name": "rollback_test", "method": "DELETE", "route": "/test"}
                ],
            },
            {"method": "GET", "route": "/test"},
        ],
    )

    for _ in range(3):
        result = loop.run_until_complete(manager.next())

    assert result["name"] == "rollback_test"
    assert result["ignore"] is True

    httpretty.disable()
    httpretty.reset()


def test_rollback_multi_urls():
    """Test spintest with rollback on multiple URLS."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://foo.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://foo.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://foo.com/test", status=204)
    httpretty.register_uri(httpretty.POST, "http://bar.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://bar.com/test", status=200)
    httpretty.register_uri(httpretty.DELETE, "http://bar.com/test", status=204)

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://foo.com", "http://bar.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
        ],
    )

    for _ in range(3):
        result = loop.run_until_complete(manager.next())

    assert result["name"] == "delete_test"
    assert result["ignore"] is True

    for _ in range(3):
        result = loop.run_until_complete(manager.next())

    assert result["name"] == "delete_test"
    assert result["ignore"] is False

    httpretty.disable()
    httpretty.reset()


def test_rollback_multi_urls_parallel():
    """Test spintest with rollback on multiple URLS parallel."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://foo.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://foo.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://foo.com/test", status=204)
    httpretty.register_uri(httpretty.POST, "http://bar.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://bar.com/test", status=200)
    httpretty.register_uri(httpretty.DELETE, "http://bar.com/test", status=204)

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://foo.com", "http://bar.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
        ],
        parallel=True,
    )

    for _ in range(4):
        result = loop.run_until_complete(manager.next())

    assert result["name"] == "delete_test"
    assert result["ignore"] is True

    httpretty.disable()
    httpretty.reset()


def test_rollback_multi_rollbacks():
    """Test spintest with multiple rollbacks."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test_2", status=204)

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://test.com"],
        [
            {
                "method": "POST",
                "route": "/test",
                "rollback": ["delete_test", "delete_test_2"],
            },
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
            {"name": "delete_test_2", "method": "DELETE", "route": "/test_2"},
        ],
    )

    for _ in range(4):
        result = loop.run_until_complete(manager.next())

    assert result["name"] == "delete_test_2"
    assert result["ignore"] is True

    httpretty.disable()
    httpretty.reset()


def test_rollback_multi_rollbacks_parallel():
    """Test spintest with multiple rollbacks parallel."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test_2", status=204)

    loop = asyncio.new_event_loop()
    manager = TaskManager(
        ["http://test.com"],
        [
            {
                "method": "POST",
                "route": "/test",
                "rollback": ["delete_test", "delete_test_2"],
            },
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
            {"name": "delete_test_2", "method": "DELETE", "route": "/test_2"},
        ],
    )

    for _ in range(4):
        result = loop.run_until_complete(manager.next())

    assert result["name"] == "delete_test_2"
    assert result["ignore"] is True

    httpretty.disable()
    httpretty.reset()


def test_rollback_templating():
    """Test spintest with templating."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.POST, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/bar", status=204)

    spintest(
        ["http://test.com"],
        [
            {
                "method": "POST",
                "route": "/test",
                "output": "test",
                "rollback": ["delete_test"],
            },
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/{{ test['foo'] }}"},
        ],
    )

    assert httpretty.last_request().method == httpretty.DELETE

    httpretty.disable()
    httpretty.reset()


def test_rollback_templating_definition_inside():
    """Test spintest with templating inside rollback."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=500)
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"foo": "bar"}),
        status=200,
    )
    httpretty.register_uri(httpretty.DELETE, "http://test.com/bar", status=204)

    spintest(
        ["http://test.com"],
        [
            {
                "method": "POST",
                "route": "/test",
                "rollback": [
                    {"method": "GET", "route": "/test", "output": "test"},
                    {"method": "DELETE", "route": "/{{ test['foo'] }}"},
                ],
            }
        ],
    )

    assert httpretty.last_request().method == httpretty.DELETE

    httpretty.disable()
    httpretty.reset()


def test_rollback_templating_definition_inside_parallel():
    """Test spintest with templating inside rollback."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=500)
    httpretty.register_uri(
        httpretty.GET,
        "http://test.com/test",
        body=json.dumps({"foo": "bar"}),
        status=200,
    )
    httpretty.register_uri(httpretty.DELETE, "http://test.com/bar", status=204)

    spintest(
        ["http://test.com"],
        [
            {
                "method": "POST",
                "route": "/test",
                "rollback": [
                    {"method": "GET", "route": "/test", "output": "test"},
                    {"method": "DELETE", "route": "/{{ test['foo'] }}"},
                ],
            }
        ],
        parallel=True,
    )

    assert httpretty.last_request().method == httpretty.DELETE

    httpretty.disable()
    httpretty.reset()


def test_rollback_templating_parallel():
    """Test spintest with templating parallel."""
    httpretty.enable()
    httpretty.register_uri(
        httpretty.POST, "http://test.com/test", body=json.dumps({"foo": "bar"})
    )
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/bar", status=204)

    spintest(
        ["http://test.com"],
        [
            {
                "method": "POST",
                "route": "/test",
                "output": "test",
                "rollback": ["delete_test"],
            },
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/{{ test['foo'] }}"},
        ],
        parallel=True,
    )

    assert httpretty.last_request().method == httpretty.DELETE

    httpretty.disable()
    httpretty.reset()


def test_rollback_failed():
    """Test spintest with failed rollback."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test", status=500)

    spintest(
        ["http://test.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
        ],
    )

    assert httpretty.last_request().method == httpretty.DELETE

    httpretty.disable()
    httpretty.reset()


def test_rollback_failed_parallel():
    """Test spintest with failed rollback."""
    httpretty.enable()
    httpretty.register_uri(httpretty.POST, "http://test.com/test", status=201)
    httpretty.register_uri(httpretty.GET, "http://test.com/test", status=500)
    httpretty.register_uri(httpretty.DELETE, "http://test.com/test", status=500)

    spintest(
        ["http://test.com"],
        [
            {"method": "POST", "route": "/test", "rollback": ["delete_test"]},
            {"method": "GET", "route": "/test"},
            {"name": "delete_test", "method": "DELETE", "route": "/test"},
        ],
        parallel=True,
    )

    assert httpretty.last_request().method == httpretty.DELETE

    httpretty.disable()
    httpretty.reset()
